/******************************************************************

	MUSCLE SmartCard Development ( http://www.linuxnet.com )
	    Title  : dyn_unix.c
	    Package: pcsc lite
            Author : David Corcoran
            Date   : 8/12/99
            Purpose: This abstracts dynamic library loading 
                     functions and timing. 

********************************************************************/

#include <dyn_generic.h>
#include <debuglog.h>
#include <string.h>
#include <dlfcn.h>

int DYN_LoadLibrary( void** pvLHandle, char* pcLibrary ) {

  char *error;

  *pvLHandle = 0;
  *pvLHandle = dlopen( pcLibrary, RTLD_LAZY );

  if ( (error = dlerror()) != 0 ) {
    DebugLogC("DYN_LoadLibrary: dlerror() reports %s\n", error, __FILE__,
               __LINE__ );
    return SCARD_F_UNKNOWN_ERROR;
  }

  return SCARD_S_SUCCESS;
}

int DYN_CloseLibrary( void** pvLHandle ) {

  char *error;

  dlclose( *pvLHandle );
  *pvLHandle = 0;

  if ( (error = dlerror()) != 0 ) {
    DebugLogC("DYN_CloseLibrary: dlerror() reports %s\n", error, __FILE__,
               __LINE__ );
    return SCARD_F_UNKNOWN_ERROR;
  }

  return SCARD_S_SUCCESS;
}

int DYN_GetAddress( void* pvLHandle, void** pvFHandle, char* pcFunction ) {

  char *error;
  char *pcFunctionName;
  int iSize;

#ifdef __BSD__
  iSize             = strlen(pcFunction) + 2; /* 1-NULL, 2-_ */
  pcFunctionName    = (char *)malloc(iSize*sizeof(char));
  pcFunctionName[0] = '_';
  strcpy(&pcFunctionName[1], pcFunction);
#else
  pcFunctionName = pcFunction;
#endif

  *pvFHandle = 0;
  *pvFHandle = dlsym( pvLHandle, pcFunctionName );

  if ( (error = dlerror()) != 0 ) {
    DebugLogC("DYN_GetAddress: dlerror() reports %s\n", error, __FILE__,
               __LINE__ );
    return SCARD_F_UNKNOWN_ERROR;
  }

  return SCARD_S_SUCCESS;
}

